#ifndef BUFFER_H
#define BUFFER_H
#include <Windows.h>
#include <cstdio>
#include <new>
#include "Unit.h"

const double DEFAULT_SIZE_MULTIPLIER = 1.2; // +20%
class buffer
{
	public:

/* Constructors */	buffer();
					buffer(const char* bufferTitle);

/*  Destructor  */	~buffer();

/*  Accessors   */	ULONG size(); // Returns sizeUsed
					char* title(); // Returns title of buffer

					// Puts data in 'dest' if 'location' is within the buffer
					template <typename valueType>
						bool get(valueType &dest, ULONG location);
					bool getBit(bool &dest, ULONG location, ULONG bitNum);
					bool getString(char* dest, ULONG location, ULONG size);

					// Use the data pointer with care, don't read past dest[sizeUsed-location-1]
					void* getPtr(ULONG location);
					void* getPtr(ULONG location, ULONG sizeRequested);
					//bool getPtr(char* &dest, ULONG location);
					template <typename valueType>
						bool getPtr(valueType* &dest, ULONG location, ULONG sizeRequested);

					// Returns data if the location is < sizeUsed, otherwise returns 0
					template <typename valueType>
						valueType get(ULONG location);
					bool getBit(ULONG location, ULONG bitNum);

/*   Searching   */	// Returns true if the sequence was found at the location
					bool has(char character, ULONG location);
					bool has(const char* str, ULONG location, ULONG size);

					// Puts next instance of 'character' in 'dest' if found
					bool getNext(char character, ULONG start, ULONG &dest);

/*   Mutators   */	// Adds data onto the end of the buffer
					template <typename valueType>
						bool add(valueType value);
					template <typename valueType>
						bool add(valueType value, ULONG amount);
					bool addStr(const char* chunk, ULONG chunkSize);

					// Inserts data at a position in the buffer
					template <typename valueType>
						bool insert(ULONG location, valueType value);
					bool insertStr(ULONG startLocation, const char* chunk, ULONG chunkSize);

					// Replaces the specified portion of the buffer
					template <typename valueType>
						bool replace(ULONG location, valueType value);
					bool replaceStr(ULONG startLocation, const char* chunk, ULONG chunkSize);
					bool replaceStr(ULONG startLocation, ULONG initialSize, const char* chunk, ULONG chunkSize);

					// Replace the whole buffer
					bool overwrite(const char* chunk, ULONG chunkSize);

					// Deletes a part of the buffer
					template <typename valueType>
						bool del(ULONG location);
					bool del(ULONG startLocation, ULONG endLocation);

/*  IO Methods  */	// Writes the data to 'pFile'
					void write(FILE* pFile);

					// Sets the title to the first four characters and extracts data the size of the following long
					bool extract(buffer &buf, ULONG &position);

					friend bool FileToBuffer(HANDLE &hMpq, const char* fileName, buffer &buf);
					friend bool FileToBuffer(const char* FileName, buffer &buf);
					friend void GetConfig();

/*     Misc     */  // Sets the size of the buffer
					bool setSize(ULONG size);

					// Deletes data and sets sizeUsed and sizeAllotted to zero
					void flush();


	private:

/*	   Data	    */  char* data; // The actual data contained by the buffer
					char bufTitle[5]; // Title of the buffer
					ULONG sizeAllotted; // Size allocated for data
					ULONG sizeUsed; // Size of data being used

/* Priv Methods */  // Called to increase or decrease buffer size
					bool resize(signed long long sizeChange, bool multiplySize);

};

class BadResize : public std::bad_alloc
{
	public:
		explicit BadResize(const char* message = "") : std::bad_alloc( message ) { }
		virtual ~BadResize() throw() { }
};

#endif