#include <Windows.h>
#include <CommCtrl.h>
#include <tchar.h>
#include <tlhelp32.h>
#include <fstream>
#include <iostream>
#include <time.h>
#include <string>

#include "SFmpqapi.h"
#include "Functions.h"
#include "Buffer.h"
#include "DefaultMap.h"
#include "Map.h"
#include "Math.h"
#include "Debug.h"
#include "FileIO.h"

using namespace std;

const int STOP_BUTTON = 100;
const int GO_BUTTON = 101;
const int UNIT_ID = 102;
const int PLAYER = 103;
const int LOAD = 104;

HWND hMain,
	 hSC,
	 hEdit, hText, hEdit2, hEdit3, hText2, hText3, hButton, hButton2;

PROCESS_INFORMATION PI = { };
MAP map;

char logFileName[MAX_PATH];

COLORREF cBackground = RGB(201, 211, 226);
HBRUSH hBackground = CreateSolidBrush( cBackground );

bool running = true;
int LoadTime = 8000;
char SCPath[MAX_PATH];

UINT step = 0;
USHORT unitID = 227;
UCHAR playerMin = 0, playerMax = 0;
UCHAR player = 2;
FILE *pFile;

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);	// For the main window
void DoTest();
void CloseSC();
void GetConfig();

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	AllocConsole();
	freopen("CONOUT$", "w", stdout);

	WNDCLASSEX wcMainWin = { };

	wcMainWin.cbSize		 = sizeof(WNDCLASSEX);
    wcMainWin.lpfnWndProc    = WndProc;
    wcMainWin.hInstance      = hInstance;
    wcMainWin.hCursor        = LoadCursor(NULL, IDC_ARROW);
	wcMainWin.hbrBackground  = hBackground;
    wcMainWin.lpszClassName  = "MainWin";

	if ( !RegisterClassEx(&wcMainWin) )
	{
		MessageBox(NULL, "Window Registration Failed!", "Error!", MB_ICONEXCLAMATION | MB_OK);
		return 0;
	}

	hMain = CreateWindowEx(0, "MainWin", "EUP Tester v1.0",
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
		CW_USEDEFAULT, CW_USEDEFAULT, 250, 250,
		NULL, NULL, hInstance, NULL);

	if ( !hMain )
	{
		MessageBox(NULL, "Window Creation Failed!", "Error!", MB_ICONEXCLAMATION | MB_OK);
		return 1;
	}

	if ( !RegisterHotKey(hMain, 0, 0, VK_F4) )
		MessageBox(NULL, "Hotkey Registration Failed!", NULL, MB_OK);

	MpqInitialize();

	InitCommonControls();
	ShowWindow(hMain, SW_SHOW);
    UpdateWindow(hMain);

	CloseSC();

	map.CreateNew(256, 256, 0, 0, 0);

	GetConfig();

	map.SaveType = 3;
	map.SaveFile(hMain, false);

	player = playerMin;
	
	HWND hSC = NULL;

	MSG msg;
	ZeroTimer();

	RECT rcMain;
	GetWindowRect(hMain, &rcMain);

	SetWindowPos(GetConsoleWindow(), NULL, rcMain.right, rcMain.top, rcMain.right, rcMain.bottom, NULL); 

	while ( running )
	{
		if ( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) )
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		DoTest();
		Sleep(1); // No need to consume a core
	}
	return msg.wParam;
}

void StopTest()
{
	ReleaseKey(VK_RETURN);
	fclose(pFile);
	cout << endl;
	step = 0;
	if ( PI.hProcess != NULL )
	{
		TerminateProcess(PI.hProcess, 0);
		WaitForSingleObject(PI.hProcess, INFINITE);
	}
	SendMessage(hEdit, EM_SETREADONLY, false, NULL);
	SendMessage(hEdit2, EM_SETREADONLY, false, NULL);
	SendMessage(hEdit3, EM_SETREADONLY, false, NULL);
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch ( msg )
	{
		case WM_HOTKEY:
			if ( HIWORD(lParam) == VK_F4 )
			{
				StopTest();
			}
			break;

		case WM_CREATE:
			{
				hEdit = CreateWindowEx(WS_EX_CLIENTEDGE, TEXT("Edit"), TEXT("227"),
					WS_CHILD | WS_VISIBLE,
					120, 21, 80, 20,
					hWnd, NULL, NULL, NULL);
				
				hText = CreateWindowEx(NULL, TEXT("Static"), TEXT("First UnitID:"),
					WS_VISIBLE | WS_CHILD,
					40, 20, 80, 20,
					hWnd, (HMENU)UNIT_ID, NULL, NULL);

				hEdit2 = CreateWindowEx(WS_EX_CLIENTEDGE, TEXT("Edit"), TEXT("9-9"),
					WS_CHILD | WS_VISIBLE,
					120, 51, 80, 20,
					hWnd, NULL, NULL, NULL);

				hText2 = CreateWindowEx(NULL, TEXT("Static"), TEXT("Player Range:"),
					WS_VISIBLE | WS_CHILD,
					25, 50, 95, 20,
					hWnd, (HMENU)PLAYER, NULL, NULL);

				hEdit3 = CreateWindowEx(WS_EX_CLIENTEDGE, TEXT("Edit"), TEXT("8000"),
					WS_CHILD | WS_VISIBLE,
					120, 81, 80, 20,
					hWnd, NULL, NULL, NULL);

				hText3 = CreateWindowEx(NULL, TEXT("Static"), TEXT("Map-Load Time:"),
					WS_VISIBLE | WS_CHILD,
					15, 80, 105, 20,
					hWnd, (HMENU)LOAD, NULL, NULL);

				hButton = CreateWindowEx(NULL, TEXT("Button"), TEXT("Go"),
					WS_VISIBLE | WS_CHILD,
					40, 110, 80, 40,
					hWnd, (HMENU)GO_BUTTON, NULL, NULL);

				hButton2 = CreateWindowEx(NULL, TEXT("Button"), TEXT("Stop (F4)"),
					WS_VISIBLE | WS_CHILD,
					125, 110, 80, 40,
					hWnd, (HMENU)STOP_BUTTON, NULL, NULL);
			}
			break;

		case WM_CTLCOLORSTATIC:
		{
			HDC hdcStatic = (HDC)wParam;
			SetTextColor(hdcStatic, RGB(0, 0, 0));
			SetBkColor(hdcStatic, cBackground);
			return (INT_PTR)CreateSolidBrush(cBackground);
		}
		break;

		case WM_COMMAND:
			switch LOWORD(wParam)
			{
				case GO_BUTTON:
					{
						SendMessage(hEdit, EM_SETREADONLY, true, NULL);
						SendMessage(hEdit2, EM_SETREADONLY, true, NULL);
						SendMessage(hEdit3, EM_SETREADONLY, true, NULL);

						char cText[128];
						
						int length = GetWindowText(hEdit, cText, 40);
						cText[length] = NULL;

						unitID = atoi(cText);

						length = GetWindowText(hEdit2, cText, 120);
						cText[length] = NULL;

						int dashPos = 0;
						while ( cText[dashPos] != '-' && dashPos < 127 )
							dashPos++;

						char cFirst[128];
						strcpy(cFirst, cText);
						cFirst[dashPos] = NULL;
						playerMin = atoi(cFirst)-1;
						player = playerMin;

						char cLast[128];
						strcpy(cLast, &cText[dashPos+1]);
						playerMax = atoi(cLast)-1;

						length = GetWindowText(hEdit3, cText, 40);
						cText[length] = NULL;

						LoadTime = atoi(cText);

						int logNum = 0;
						do
						{
							if ( pFile )
								fclose(pFile);
							sprintf_s(logFileName, "EupLog - %i.txt", logNum);
							logNum ++;
						}
						while ( fopen(logFileName, "r") != nullptr );

						pFile = fopen(logFileName, "w");
						if ( !pFile )
						{
							MessageBox(NULL, "Couldn't create logfile", NULL, MB_OK);
							return 0;
						}
						else
							cout << "Logging to: " << logFileName << endl << endl;

						step = 1;
					}
					break;
				
				case STOP_BUTTON:
					StopTest();
					break;
			}
			break;

		case WM_DESTROY:
			StopTest();
			running = false;
			PostQuitMessage(0);
			break;

		default:
			return DefWindowProc(hWnd, msg, wParam, lParam);
			break;
	}
	return 0;
}

void DoTest()
{
	/** 
		__ Steps __

		0 : NULL
		1 : Edit map, increment player and/or unitID
		2 : Launch StarCraft
		3 : Open multi-player menu ('M') and expansion ('E')
		4 : Arrow down to LAN UPD (4x 'Arrow Down') and Enter LAN UPD ('Enter')
		5 : Choose first name ('Enter')
		6 : Choose Create Game ('G')
		7 : Choose test map ('Enter')
		8 : Launch ('Alt+O')
		9 : Skip briefing
		10: Check for crash
		11: Check memory for unitID, record result
		12: Open Menu ('F10')
		13: End Mission ('E')
		14: Surrender ('S')
		15: Quit Mission ('Q')
		16: Exit High Scores ('O')
	*/

	switch ( step )
	{
		case 0:
			// Do Nothing! Can be used for debug/pausing.
			return;
			break;

		case 1: // Edit map
			ZeroTimer();
			ZeroTimer2();
			if ( hSC == NULL )
			{
				if ( player == playerMax )
				{
					// Update UnitID Text
						char cID[32] = { };
						itoa(unitID, cID, 10);
						SetWindowText(hEdit, cID);

					map.UNIT->replace<USHORT>(8, unitID); // Set UnitID

					player = playerMin;
					map.UNIT->replace<BYTE>(16, player); // Set Player
					unitID ++;
				}
				else if ( player < playerMax )
				{
					map.UNIT->replace<USHORT>(8, unitID); // Set UnitID
					map.UNIT->replace<BYTE>(16, player); // Set Player
					player++;
				}
				if ( map.SaveFile(hMain, false) )
					step ++;
				else
					Sleep(50);
			}
			break;

		case 2: // Create StarCraft Window
			{
				STARTUPINFO SUI = { };
				SUI.cb = sizeof(STARTUPINFO);

				if ( CreateProcess(SCPath, NULL, NULL, NULL, false, NULL, NULL, NULL, &SUI, &PI) )
				{
					ZeroTimer();
					step ++;
				}
				else
					Sleep(50);
			}
			break;

		case 3: // Select Multiplayer-Expansion
			if ( Timer() > 10 )
			{
				UINT track = 0;
				SendKey('M');
				ReadMem(0x6CE0E4, track);

				if ( track == 11 ) // Detects multiplayer menu open
				{
					SendKey('E');
					ZeroTimer();
					step ++;
				}
				else
					ZeroTimer();
			}
			break;

		case 4: // Select LAN UPD
			if ( Timer() > 10 )
			{
				UINT track = 0;

				ReadMem(0x6D5A28, track);

				if ( track == 1112425812 )
					SendKey(VK_NEXT);

				ReadMem(0x6D5A28, track);

				if ( track == 1430540366 )
					SendKey(VK_UP);
				
				ReadMem(0x6D5A28, track);

				if ( track == 1396916812 )
				{
					SendKey(VK_RETURN);
					step++;
				}
				ZeroTimer();
			}
			break;

		case 5: // Select Name
			if ( Timer() > 10 )
			{
				UINT track = 0;
				SendKey(VK_RETURN);
				ReadMem(0x006D0534, track);

				if ( track == 15 )
					step++;

				ZeroTimer();
			}
			break;

		case 6: // Create Game
			if ( Timer() > 10 )
			{
				SendKey('G');

				UINT track = 0;
				ReadMem(0x006D0534, track);

				if ( track == 38 )
					step++;

				ZeroTimer();
			}
			break;

		case 7: // Select Map
			if ( Timer() > 10 )
			{
				SendKey(VK_RETURN);

				UINT track = 0;
				ReadMem(0x006D0534, track);

				if ( track == 84 )
					step++;

				ZeroTimer();
			}
			break;

		case 8: // Start Game
			if ( Timer() > 10 )
			{
				PressKey(VK_MENU);
				SendKey('O');
				ReleaseKey(VK_MENU);

				UINT track = 0;
				ReadMem(0x0066FBFC, track);

				if ( track > 0 )
				{
					WriteMem(0x0066FBFC, 1);
					step ++;
				}

				ZeroTimer();
			}
			break;

		case 9: // Skip Briefing
			{
				UINT track = 0;
				ReadMem(0x0066FBFC, track);

				if ( track == 0 )
				{
					PressKey(VK_RETURN);
					ZeroTimer();
					step++;
				}
			}
			break;

		case 10: // Crash Check
			if ( Timer() > LoadTime )
			{
				ReleaseKey(VK_RETURN);
				UINT track = 0;
				if ( ReadMem(0x57F180, track) ) // If memory is readable and minerals > 0
				{
					if ( track > 0 )
					{
						WriteMem(0x57F180, 0); // Set minerals to 0
						step++; // No crash, continue
						break;
					}
				}

				// Crashed, force-quit starcraft, start next test
				TerminateProcess(PI.hProcess, 0);
				WaitForSingleObject(PI.hProcess, INFINITE);
				SendKey(VK_RIGHT);
				SendKey(VK_RETURN);
				step = 1;
			}
			break;

		case 11: // Check that unit was placed
			{
				UINT indexZeroUnitID;
				ReadMem(0x59CD0C, indexZeroUnitID);
				if ( indexZeroUnitID == unitID-1 || (player == playerMin && indexZeroUnitID == unitID-1) )
				{
					// Record Success
					char str[256] = { };
					sprintf(str, "UnitID: %u	Player: %u\n", indexZeroUnitID, (USHORT)player+1);
					cout << str;
					fwrite(str, strlen(str), 1, pFile);
				}

				TerminateProcess(PI.hProcess, 0);
				WaitForSingleObject(PI.hProcess, INFINITE);
				step = 1;
			}
			break;

		case 12:
			step++;
			break;

		case 13:
			step++;
			break;

		case 14:
			step++;
			break;

		case 15:
			step++;
			break;

		case 16:
			step++;
			break;

		case 17:
			step++;
			break;
	}

	if ( Timer2() > 30000 ) // If at any time you've been waiting for 30 seconds
	{
		TerminateProcess(PI.hProcess, 0);
		step = 1;
	}
}

void CloseSC()
{
	hSC = FindWindow(NULL, "Brood War");

	if ( hSC != NULL )
	{
		if ( MessageBox(NULL, "EupTest cannot start until StarCraft is closed.\n\n Would you like EupTest to close StarCraft?", NULL, MB_YESNO) == IDYES )
		{
			DWORD PID;
			GetWindowThreadProcessId(hSC, &PID);
			HANDLE pHandle = OpenProcess(PROCESS_TERMINATE, false, PID);
			TerminateProcess(pHandle, 0);
			CloseHandle(pHandle);
			hSC = NULL;
		}
		else
			CloseSC();
	}
}

void GetConfig()
{
	char starcraftDir[MAX_PATH];
	GetScPath(starcraftDir);
	sprintf(map.filename, "%s\\Maps\\BroodWar\\!!!.scx", starcraftDir);
	sprintf(SCPath, "%s\\StarCraft.exe", starcraftDir);
}