#include "FileIO.h"
#include "SFmpqapi.h"
#include "Debug.h"
#include <sys/stat.h>

bool OpenArchive(const char* directory, const char* fileName, MPQHANDLE &hMpq)
{
	struct stat buf;
	char filePath[MAX_PATH] = { };

	if ( directory != NULL )
		sprintf_s(filePath, "%s\\", directory);
	strcat_s(filePath, fileName);

	char locateError[MAX_PATH+65] = { };
	sprintf_s(locateError, "Could not find %s!\n\nWould you like to locate it manually?", filePath);

	char openFilter[MAX_PATH+40] = { };
	sprintf_s(openFilter, "%s\0*%s*;\0All Files\0*.*\0\0", fileName, fileName);

	char retryError[MAX_PATH+70] = "Failed to open ";
	sprintf_s(retryError, "Failed to open %s! The file may be in use.\n\nWould you like to try again?", fileName);

	if ( stat(filePath, &buf) == -1 )
	{
		if ( MessageBox(NULL, locateError, "Error!", MB_YESNO|MB_ICONEXCLAMATION) == IDYES )
		{
			OPENFILENAME ofn = { };
	
			ofn.lStructSize = sizeof(ofn);
			ofn.hwndOwner = NULL;
			ofn.lpstrFilter = openFilter;
			ofn.lpstrFile = filePath;
			ofn.nMaxFile = MAX_PATH;
			ofn.Flags = OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT;
	
			if ( GetOpenFileName(&ofn) )
			{
				if ( SFileOpenArchive(filePath, 0, NULL, &hMpq) )
				{
					if ( hMpq == NULL )
						return false;
					else
						return true;
				}
				else
					return false;
			}
		}
		else
			return false;
	}
	else
	{
		if ( SFileOpenArchive(filePath, 0, NULL, &hMpq) )
			return true;
		else if ( !(MessageBox(NULL, retryError, "Error!", MB_YESNO|MB_ICONEXCLAMATION) == IDYES) )
			return false;
	}
	return OpenArchive(directory, fileName, hMpq);
}

bool FileToBuffer(HANDLE &hMpq, const char* fileName, buffer &buf)
{
	if ( hMpq == nullptr )
		sprintf_s(LastError, MAX_ERROR_LENGTH, "NULL MPQ file specified for opening %s", fileName);
	else
	{
		ULONG bytesRead = 0;
		HANDLE openFile = NULL;
		if ( SFileGetFileInfo(hMpq, SFILE_INFO_NUM_FILES) != 0xFFFFFFFF )
		{
			if ( SFileOpenFileEx(hMpq, fileName, SFILE_SEARCH_CURRENT_ONLY, &openFile) )
			{
				unsigned long fileSize = SFileGetFileSize(openFile, NULL);
				if ( buf.setSize(fileSize) )
				{
					buf.sizeUsed = fileSize;
					SFileReadFile(openFile, (LPVOID)buf.data, buf.sizeUsed, &bytesRead, NULL);

					if ( buf.sizeUsed == bytesRead )
						return true;
				}
			}
			else
				sprintf_s(LastError, MAX_ERROR_LENGTH, "Failed to get %s from MPQ file", fileName);
		}
		else
			sprintf_s(LastError, MAX_ERROR_LENGTH, "File is already open", fileName);
	}
	return false;
}

bool FileToBuffer(HANDLE &hStarDat, HANDLE &hBrooDat, HANDLE &hPatchRt, const char* fileName, buffer &buf)
{
	return	  FileToBuffer(hPatchRt, fileName, buf)
		   || FileToBuffer(hBrooDat, fileName, buf)
		   || FileToBuffer(hStarDat, fileName, buf);
}

bool FileToBuffer(const char* FileName, buffer &buf)
{
	FILE* pFile;
	fopen_s(&pFile, (const char*)FileName, "rb");
	
	if ( pFile == nullptr )
		return false;
	
	buf.flush();
	fseek (pFile , 0, SEEK_END);
	unsigned long fileSize = ftell(pFile);

	if ( !buf.setSize(fileSize) )
		return false;

	buf.sizeUsed = fileSize;
	rewind (pFile);

	size_t lengthRead = fread(buf.data, 1, buf.sizeUsed, pFile);
	fclose(pFile);

	if ( lengthRead != buf.sizeUsed )
		return false;

	return true;
}

DWORD GetSubKeyString(HKEY hParentKey, const char* lpSubKey, const char* lpValueName, char* lpData, DWORD* lpcbData)
{
	HKEY hKey = NULL;

	DWORD errorCode = RegOpenKeyEx(hParentKey, lpSubKey, 0, KEY_QUERY_VALUE, &hKey); // Open key
	if ( errorCode != ERROR_SUCCESS ) // Open key failed
		return errorCode; // Return error code

	errorCode = RegQueryValueEx(hKey, lpValueName, NULL, NULL, (LPBYTE)lpData, lpcbData); // Read key

	RegCloseKey(hKey); // Close key

	return errorCode; // Return success/error message
}

void GetRegString(char* dest, const char* lpSubKey, const char* lpValueName)
{
	DWORD lpcbData = MAX_PATH;

	DWORD errorCode = GetSubKeyString(HKEY_CURRENT_USER, lpSubKey, lpValueName, dest, &lpcbData); // Try HKCU

	if ( errorCode != ERROR_SUCCESS ) // Not found in HKCU
		GetSubKeyString(HKEY_LOCAL_MACHINE, lpSubKey, lpValueName, dest, &lpcbData); // Try HKLM
}

void GetScPath(char* dest)
{
	GetRegString(dest, "SOFTWARE\\Blizzard Entertainment\\Starcraft", "InstallPath");
}