#include "Map.h"
#include "Resource.h"
#include "Math.h"
#include "DefaultMap.h"
#include "SFmpqapi.h"
#include "FileIO.h"
#include "GuiInfo.h"
#include "Debug.h"

MAP::MAP() : lpvBits(nullptr), bitHeight(0), bitWidth(0), hWnd(NULL), SaveType(0),
			 layer(0), player(0), zoom(1), RedrawMiniMap(true), RedrawMap(true),
			 dragging(false), isProtected(false), head(nullptr),
			 TYPE(nullptr),	VER (nullptr),	IVER(nullptr),	IVE2(nullptr),
			 VCOD(nullptr),	IOWN(nullptr),	OWNR(nullptr),	ERA (nullptr),
			 DIM (nullptr),	SIDE(nullptr),	MTXM(nullptr),	PUNI(nullptr),
			 UPGR(nullptr),	PTEC(nullptr),	UNIT(nullptr),	ISOM(nullptr),
			 TILE(nullptr),	DD2 (nullptr),	THG2(nullptr),	MASK(nullptr),
			 STR (nullptr),	UPRP(nullptr),	UPUS(nullptr),	MRGN(nullptr),
			 TRIG(nullptr),	MBRF(nullptr),	SPRP(nullptr),	FORC(nullptr),
			 WAV (nullptr),	UNIS(nullptr),	UPGS(nullptr),	TECS(nullptr),
			 SWNM(nullptr),	COLR(nullptr),	PUPx(nullptr),	PTEx(nullptr),
			 UNIx(nullptr),	UPGx(nullptr),	TECx(nullptr)
{
	filename[0] = NULL;
	display.x = 0;
	display.y = 0;
	grids[0].size.x = 0; grids[0].size.y = 0;
}

MAP::~MAP()
{
	if ( lpvBits != nullptr )
	{
		delete[] lpvBits;
		lpvBits = nullptr;
	}

	section* next;
	while ( head != nullptr )
	{
		next = head->next;
		delete head;
		head = next;
	}

	DeleteObject(MemBitmap);
	DeleteObject(MemMiniBitmap);
	DeleteDC(MemhDC);
	DeleteDC(MemMinihDC);
}

bool MAP::RemoveSection(buffer* buf)
{
	section* curr = head,
		   * prev = nullptr;

	while ( curr )
	{
		if ( &curr->buf == buf )
		{
			if ( prev )
				prev->next = curr->next;
			else
				head = curr->next;

			delete curr;
			curr = nullptr;
			UpdatePtrs();
			return true;
		}
		prev = curr;
		curr = curr->next;
	}
	return false;
}

void MAP::UpdatePtrs()
{
	section* curr = head;
	char* header = nullptr;

	TYPE = VER  = IVER = IVE2 = VCOD = IOWN = DIM  = SIDE = MTXM = PUNI = UPGR = PTEC = UNIT = ISOM =
	TILE = DD2  = THG2 = MASK = STR  = UPRP = UPUS = MRGN = TRIG = MBRF = SPRP = FORC = WAV  = UNIS =
	UPGS = TECS = SWNM = COLR = PUPx = PTEx = UNIx = UPGx = TECx = nullptr;

	while ( curr != nullptr )
	{
		header = curr->buf.title();
		if		( !strcmp(header, "TYPE") ) TYPE = &curr->buf; else if ( !strcmp(header, "VER ") ) VER  = &curr->buf;
		else if ( !strcmp(header, "IVER") ) IVER = &curr->buf; else if ( !strcmp(header, "IVE2") ) IVE2 = &curr->buf;
		else if ( !strcmp(header, "VCOD") ) VCOD = &curr->buf; else if ( !strcmp(header, "IOWN") ) IOWN = &curr->buf;
		else if ( !strcmp(header, "OWNR") ) OWNR = &curr->buf; else if ( !strcmp(header, "ERA ") ) ERA  = &curr->buf;
		else if ( !strcmp(header, "DIM ") ) DIM  = &curr->buf; else if ( !strcmp(header, "SIDE") ) SIDE = &curr->buf;
		else if ( !strcmp(header, "MTXM") ) MTXM = &curr->buf; else if ( !strcmp(header, "PUNI") ) PUNI = &curr->buf;
		else if ( !strcmp(header, "UPGR") ) UPGR = &curr->buf; else if ( !strcmp(header, "PTEC") ) PTEC = &curr->buf;
		else if ( !strcmp(header, "UNIT") ) UNIT = &curr->buf; else if ( !strcmp(header, "ISOM") ) ISOM = &curr->buf;
		else if ( !strcmp(header, "TILE") ) TILE = &curr->buf; else if ( !strcmp(header, "DD2 ") ) DD2  = &curr->buf;
		else if ( !strcmp(header, "THG2") ) THG2 = &curr->buf; else if ( !strcmp(header, "MASK") ) MASK = &curr->buf;
		else if ( !strcmp(header, "STR ") ) STR  = &curr->buf; else if ( !strcmp(header, "UPRP") ) UPRP = &curr->buf;
		else if ( !strcmp(header, "UPUS") ) UPUS = &curr->buf; else if ( !strcmp(header, "MRGN") ) MRGN = &curr->buf;
		else if ( !strcmp(header, "TRIG") ) TRIG = &curr->buf; else if ( !strcmp(header, "MBRF") ) MBRF = &curr->buf;
		else if ( !strcmp(header, "SPRP") ) SPRP = &curr->buf; else if ( !strcmp(header, "FORC") ) FORC = &curr->buf;
		else if ( !strcmp(header, "WAV ") ) WAV  = &curr->buf; else if ( !strcmp(header, "UNIS") ) UNIS = &curr->buf;
		else if ( !strcmp(header, "UPGS") ) UPGS = &curr->buf; else if ( !strcmp(header, "TECS") ) TECS = &curr->buf;
		else if ( !strcmp(header, "SWNM") ) SWNM = &curr->buf; else if ( !strcmp(header, "COLR") ) COLR = &curr->buf;
		else if ( !strcmp(header, "PUPx") ) PUPx = &curr->buf; else if ( !strcmp(header, "PTEx") ) PTEx = &curr->buf;
		else if ( !strcmp(header, "UNIx") ) UNIx = &curr->buf; else if ( !strcmp(header, "UPGx") ) UPGx = &curr->buf;
		else if ( !strcmp(header, "TECx") ) TECx = &curr->buf;
		curr = curr->next;
	}
}

bool MAP::GoodVCOD()
{
	buffer defaultVCOD("vcod");
	Get_VCOD(defaultVCOD);

	char* defaultData = (char*)defaultVCOD.getPtr(0),
		* inData = (char*)VCOD->getPtr(0);

	int defaultSize = defaultVCOD.size();

	
	for ( int pos = 0; pos < defaultSize-8; pos++ )
	{
		if ( defaultData[pos+8] != inData[pos] )
			return false;
	}
	return true;
}

bool MAP::ParseBuffer(buffer &chk)
{
	int maxSections = 500;
	unsigned long position = 0;
	signed long size = 0;

	char header[5] = { };

	section* last,
		   * curr;
	
	while ( position+8 < chk.size() )
	{
		if ( !chk.getString(header, position, 4) )
			return false;

		section* checkExisting = head;
		while ( checkExisting != nullptr && !isProtected )
		{
			if ( !strcmp(checkExisting->buf.title(), header) ) // Duplicate section
				isProtected = true;

			checkExisting = checkExisting->next;
		}

		curr = new section(header);
		curr->next = nullptr;

		if ( head == nullptr )
			head = curr;
		else
			last->next = curr;
		
		last = curr;

		if ( !curr->buf.extract(chk, position) && chk.get<LONG>(size, position) )
		{
			if ( chk.get<LONG>(size, position) )
			{
				position += 4;
				if ( position+size > chk.size() && !isProtected ) // The size would take you past the end of the data
					isProtected = true;

				position += (signed long)size;
			}
			else // Your position was too great to get an int
				isProtected = true;
		}

		maxSections--;
		if ( !maxSections ) isProtected = true;
	}
	
	UpdatePtrs();

	if ( !GoodVCOD() )
		isProtected = true;

	if ( isProtected )
		mb("Map is protected and will be opened as view-only.");

	return true;
}

bool MAP::CreateNew(int width, int height, int tileset, int terrain, int triggers)
{
	buffer chk("nMAP");

	if (	Get_TYPE(chk) && Get_VER (chk) && Get_IVE2(chk)	&& Get_VCOD(chk)
		 && Get_IOWN(chk) && Get_OWNR(chk) && Get_ERA (chk, tileset) && Get_DIM (chk, width, height)
		 && Get_SIDE(chk) && Get_MTXM(chk, width, height) && Get_PUNI(chk) && Get_UPGR(chk)
		 && Get_UNIT(chk) && Get_PTEC(chk) && Get_ISOM(chk, width, height) && Get_TILE(chk, width, height)
		 && Get_DD2 (chk) && Get_THG2(chk) && Get_MASK(chk, width, height) && Get_STR (chk)
		 && Get_MRGN(chk, width, height)   && Get_SPRP(chk) && Get_FORC(chk) && Get_WAV (chk)
		 && Get_UNIS(chk) && Get_UPGS(chk) && Get_TECS(chk) && Get_COLR(chk)
		 && Get_PUPx(chk) && Get_PTEx(chk) && Get_UNIx(chk) && Get_UPGx(chk)
		 && Get_TECx(chk) && Get_TRIG(chk) && Get_MBRF(chk) && Get_UPRP(chk)
		 && Get_UPUS(chk) && Get_SWNM(chk) )
	{
		if ( ParseBuffer(chk) )
			return true;
		else
			Error( "New map parsing failed! Report this!" ); // Chkdraft failed to read a map it creates itself, should never happen!
	}
	return false;
}

bool MAP::LoadFile(HWND hMaps)
{
	buffer chk("oMAP");

	char* ext = strrchr(filename, '.'); // Find the last occurrence of '.'
	if ( ext != nullptr )
	{
		if ( strcmp(ext, ".scm") == 0 || strcmp(ext, ".scx") == 0 )
		{
			HANDLE hMpq = MpqOpenArchiveForUpdate(filename, MOAU_OPEN_EXISTING, 1000);

			if ( hMpq )
			{
				FileToBuffer(hMpq, "staredit\\scenario.chk", chk);
				MpqCloseUpdatedArchive(hMpq, NULL);

				if ( chk.size() > 0 && ParseBuffer(chk) )
					return true;
			}
			else if ( GetLastError() == ERROR_FILE_NOT_FOUND )
				sprintf_s(LastError, MAX_ERROR_LENGTH, "File Not Found");
			else if ( GetLastError() == MPQ_ERROR_MPQ_INVALID )
				sprintf_s(LastError, MAX_ERROR_LENGTH, "Invalid or corrupt MPQ file");
			else
				sprintf_s(LastError, MAX_ERROR_LENGTH, "%d", GetLastError());
		}
		else if ( strcmp(ext, ".chk") == 0 )
		{
			if ( FileToBuffer(filename, chk) )
			{
				if ( ParseBuffer(chk) )
					return true;
				else
					sprintf_s(LastError, MAX_ERROR_LENGTH, "Parsing Failed!");
			}
			else
				sprintf_s(LastError, MAX_ERROR_LENGTH, "Error Reading CHK File");
		}
	}
	return false;
}

bool MAP::SaveFile(HWND hMain, bool SaveAs)
{
	if ( isProtected )
		Error("Cannot save protected maps!");
	else
	{
		if ( SaveAs || !filename[0] )
		{
			OPENFILENAME ofn = GetSaveOfn(filename);
			if ( GetSaveFileName(&ofn) )
			{
				SaveType = (BYTE)ofn.nFilterIndex;

				char* ext = strrchr(filename, '.'); // Find the last occurrence of '.'
				if ( ext == nullptr )
				{
					if ( SaveType == 1 || SaveType == 2 )
						strcat_s(filename, ".scm");
					else if ( SaveType == 3 )
						strcat_s(filename, ".scx");
					else if ( SaveType >= 4 && SaveType <= 6 )
						strcat_s(filename, ".chk");
				}
				else
				{
					if ( strcmp(ext, ".chk") == 0 && SaveType < 4 )
						SaveType = 5;
					else if ( strcmp(ext, ".scm") == 0 && SaveType > 1 )
						SaveType = 2;
					else if ( strcmp(ext, ".scx") == 0 )
						SaveType = 3;
				}
			}
		}

		if ( filename[0] != NULL )
		{
			FILE* pFile(nullptr);

			if ( SaveType == 1 || SaveType == 4 ) // StarCraft Map
			{
				TYPE->overwrite("RAWS", 4);
				VER ->overwrite(";\0", 2);
				IVER->overwrite("\12\0", 2);
				IVE2->overwrite("\13\0", 2);

				if ( MRGN->size() > 1280 ) // If there's > 64 locations
					MRGN->del(1280, MRGN->size()); // Delete all over 64
			}
			else if ( SaveType == 2 || SaveType == 5 ) // Hybrid Map
			{
				TYPE->overwrite("RAWS", 4);
				VER ->overwrite("?\0", 2);
				IVER->overwrite("\12\0", 2);
				IVE2->overwrite("\13\0", 2);

				if ( MRGN->size() < 5100 ) // If there's < 255 locations
					MRGN->add<BYTE>(0, 5100-MRGN->size()); // Add till 255
			}
			else if ( SaveType == 3 || SaveType == 6 ) // BroodWar Map
			{
				VER ->overwrite("\0", 2);
				TYPE->overwrite("RAWB", 4);
				RemoveSection(IVER);
				IVE2->overwrite("\13\0", 2);

				if ( MRGN->size() < 5100 ) // If there's < 255 locations
					MRGN->add<BYTE>(0, 5100-MRGN->size()); // Add till 255
			}

			if ( SaveType > 0 && SaveType <= 3 ) // Will be packed into an MPQ
			{
				bool saved = true;
				fopen_s(&pFile, filename, "wb");
				if ( pFile != nullptr )
				{
					fclose(pFile);
					HANDLE hMpq = NULL;
					DeleteFileA("chk.tmp"); //remove any existing chk.tmp files
					fopen_s(&pFile, "chk.tmp", "wb");
					WriteFile(pFile);
					fclose(pFile);

					if ( (hMpq = MpqOpenArchiveForUpdate(filename, MOAU_OPEN_ALWAYS | MOAU_MAINTAIN_LISTFILE, 16)) )
					{
						BOOL addedFile = MpqAddFileToArchive(hMpq, "chk.tmp", "staredit\\scenario.chk", MAFA_COMPRESS | MAFA_REPLACE_EXISTING);
						MpqCloseUpdatedArchive(hMpq, NULL);

						if ( addedFile )
						{
							SetWindowText(hWnd, filename);
							DeleteFileA("chk.tmp");
							return true;
						}
						else
							Error("Failed to add file!");
					}
					else
						Error("Failed to open for update!");

					DeleteFileA("chk.tmp");
				}
				else
					Error("Failed to open file!\n\nThe file may be in use elsewhere.");
			}
			else // is a chk file or unrecognized format
			{
				DeleteFileA(filename); // Remove any existing files of the same name
				fopen_s(&pFile, filename, "wb");
				WriteFile(pFile);
				fclose(pFile);
				SetWindowText(hWnd, filename);
				return true;
			}
		}
	}
	return false;
}

void MAP::WriteFile(FILE* pFile)
{
	if ( !isProtected )
	{
		section* curr = head;
		while ( curr != nullptr )
		{
			curr->buf.write(pFile);
			curr = curr->next;
		}
	}
}

void MAP::setZoom(double newScale)
{
	zoom = newScale;
	RedrawMap = true;
	RedrawMiniMap = true;

	RECT rect;
	GetClientRect(hWnd, &rect);
	bitWidth = USHORT(((rect.right - rect.left+32)/32*32)/zoom)/32*32;
	bitHeight = USHORT(((rect.bottom - rect.top+32)/32*32)/zoom)/32*32;
	Scroll(SCROLL_X|SCROLL_Y|VALIDATE_BORDER);
	RedrawWindow(hWnd, NULL, NULL, RDW_INVALIDATE);
}

void MAP::Scroll(UINT flags)
{
	SCROLLINFO scrollbars = { };
	scrollbars.cbSize = sizeof(SCROLLINFO);
	scrollbars.fMask = SIF_ALL;
	scrollbars.nMin = 0;

	RECT rcMap;
	GetClientRect(hWnd, &rcMap);
	int screenWidth  = (rcMap.right-rcMap.left)/zoom,
		screenHeight = (rcMap.bottom-rcMap.top)/zoom;

	if ( (flags&VALIDATE_BORDER) )
		ValidateBorder(screenWidth, screenHeight);

	if ( (flags&SCROLL_X) )
	{
		scrollbars.nPos = display.x; 
		scrollbars.nMax = XSize()*32;
		scrollbars.nPage = screenWidth;
		SetScrollInfo(hWnd, SB_HORZ, &scrollbars, true);
	}
	if ( (flags&SCROLL_Y) )
	{
		scrollbars.nPos = display.y;
		scrollbars.nMax = YSize()*32;
		scrollbars.nPage = screenHeight;
		SetScrollInfo(hWnd, SB_VERT, &scrollbars, true);
	}
	RedrawMap = true;
}

void MAP::ValidateBorder(int screenWidth, int screenHeight)
{
	if ( display.x < 0 )
		display.x = 0;
	else if ( display.x > XSize()*32-screenWidth )
	{
		if ( screenWidth > XSize()*32 )
			display.x = 0;
		else
			display.x = XSize()*32-screenWidth;
	}

	if ( display.y < 0 )
		display.y = 0;
	else if ( display.y > YSize()*32-screenHeight )
	{
		if ( screenHeight > YSize()*32 )
			display.y = 0;
		else
			display.y = YSize()*32-screenHeight;
	}
}

/** ####################################
	### Begin high level map editing ###
	#################################### */

unsigned short MAP::XSize()
{
	return DIM->get<USHORT>(0);
}

unsigned short MAP::YSize()
{
	return DIM->get<USHORT>(2);
}

bool MAP::isExpansion()
{
	return VER->get<USHORT>(0) >= 205;
}

Unit* MAP::unit(unsigned short index)
{
	return (Unit*)UNIT->getPtr(index*UNIT_STRUCT_SIZE);
}

buffer* MAP::unitSettings()
{
	if ( isExpansion() )
		return this->UNIx;
	else
		return this->UNIS;
}

float MAP::Scale(unsigned short xSize, unsigned short ySize)
{
	if ( xSize >= ySize )
	{
		if ( 128.0/xSize > 1 )
			return (float)(128/xSize);
		else
			return (float)(128.0/xSize);
	}
	else
	{
		if ( 128.0/ySize > 1 )
			return (float)(128/ySize);
		else
			return (float)(128.0/ySize);
	}
}

bool MAP::SetTile(long x, long y, unsigned short tileNum)
{
	unsigned short xSize = XSize();
	if ( DIM->size() < 4 || x > xSize || y > YSize() )
		return false;

	unsigned long location = 2*xSize*y+2*x; // Down y rows, over x columns

	TILE->replace<unsigned short>(location, tileNum);
	MTXM->replace<unsigned short>(location, tileNum);

	RECT rcTile;
	rcTile.left	  = x*32-display.x;
	rcTile.right  = rcTile.left+32;
	rcTile.top	  = y*32-display.y;
	rcTile.bottom = rcTile.top+32;

	RedrawMap = true;
	RedrawMiniMap = true;
	InvalidateRect(hWnd, &rcTile, true);
	return true;
}